#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define _GNU_SOURCE
#include <getopt.h>
#include <ogg/ogg.h>

#include "xmalloc.h"
#include "stream.h"
#include "output.h"

#define VERSION_MAJOR 0
#define VERSION_MINOR 0
#define VERSION_MICRO 1

#define BUFFER_SIZE 4096

int unchain=1;
int ungroup=1;
char *outdir;

const char *broken_ogg = \
"  This Ogg file is broken and/or corrupted. This could be due to\n"
"  a flaw in the tool used to produce the file. If you believe this\n"
"  is the case, make sure to report it to the author/maintainer of\n"
"  that tool. Also note that as a consequence, some or all of the\n"
"  files procuduces by this invocation of oggsplit may also be\n"
"  invalid Ogg files.\n";

extern char *optarg;
extern int opterr, optind;
const char *optstring="cgo:uvh";
struct option options [] = {
  {"unchain", no_argument,       NULL, 'c'},
  {"ungroup", no_argument,       NULL, 'g'},
  {"outdir",  required_argument, NULL, 'o'},
  {"version", no_argument,       NULL, 'v'},
  {"help",    no_argument,       NULL, 'h'},
  {NULL,0,NULL,0}
};

static void usage(void)
{
  fprintf(stderr,
	  "Usage: oggsplit [options] inputfile [...]\n"
	  "  -c --unchain      Only split chained streams. The default is to\n"
	  "                      split both grouped and chained streams.\n"
	  "  -g --ungroup      Only split grouped streams. This will cause only\n"
	  "                      the first link to be processed.\n"
	  "  -o --outdir=dir   Put all output files in dir.\n"
	  "  -v --version      Display oggsplit version.\n"
	  "  -h --help         Display this help message.\n");
  exit(1);
}

int buffer_data(FILE *f, ogg_sync_state *oy)
{
  char *buffer=ogg_sync_buffer(oy,BUFFER_SIZE);
  int bytes=fread(buffer,1,BUFFER_SIZE,f);
  ogg_sync_wrote(oy,bytes);
  return(bytes);
}

int process_file(const char *pathname)
{
  char *dirname, *filename, *tmp;
  FILE *infile;

  output_t *junk;

  stream_ctrl_t sc;
  output_ctrl_t oc;

  ogg_sync_state oy;
  ogg_page       og;

  int group_c=0; /* a group counter */
  int chain_c=0; /* a chain counter */
  int chain_state=0; /* 0 - nothing has passed through
		      * 1 - bos pages have passed through
		      * 2 - data pages have passed through
		      */

  infile = fopen(pathname, "r");
  if(infile==NULL){
    fprintf(stderr, "Failed to open %s for reading.\n", pathname);
    return 0;
  }

  /* split out filename and dirname */
  tmp=strrchr(pathname, '/');
  if(tmp!=NULL){
    int dnlen;
    dnlen=tmp-pathname;
    dirname=xmalloc(dnlen+1);
    strncpy(dirname, pathname, dnlen);
    dirname[dnlen]='\0';

    filename=xmalloc(strlen(pathname)-dnlen);
    strcpy(filename, tmp+1);
  }else{
    dirname=".";
    filename=(char *)pathname;
  }

  stream_ctrl_init(&sc);
  output_ctrl_init(&oc, dirname, filename);

  ogg_sync_init(&oy);

  while(buffer_data(infile,&oy)){
    while(ogg_sync_pageout(&oy, &og)==1){
      stream_t *stream;

      if(ogg_page_bos(&og)){
	/* first page of stream */
	stream=stream_ctrl_stream_new(&sc, &og);

	switch(chain_state){
	case 2:
	  fprintf(stderr,
		  "WARNING: A new logical stream began before the current ones had ended.\n"
		  "  This is forbidden by the Ogg specification, and should never happen.\n\n%s",
		  broken_ogg);
	  /* note the lack of break; */
	case 0:
	  /* don't open any new links if not unchaining */
	  if(!unchain && chain_c)
	    continue;

	  /* new chain link, create new output */
	  if(unchain)chain_c++;
	  if(ungroup)group_c++;
	  stream->op=output_ctrl_output_new(&oc, chain_c, group_c);
	  chain_state=1;
	  break;
	case 1:
	  if(ungroup){
	    stream->op=output_ctrl_output_new(&oc, chain_c, ++group_c);
	  }else{
	    /* use previously opened output */
	    stream->op=&oc.outputs[oc.outputs_used-1];
	    stream->op->count++;
	  }
	  break;
	}

	printf("Ogg logical stream (serial %08x): type %s\n"
	       "  writing stream to %s\n",
	       stream->serial, stream_type_name(stream),
	       stream->op->filename);

      }else{
	/* test if this page belongs to an open stream */
	stream=stream_ctrl_stream_get(&sc, ogg_page_serialno(&og));

	if(stream==NULL){
	  /* junk in the file */
	  if(junk==NULL){
	    /* call with -1 to get the junk file */
	    junk=output_ctrl_output_new(&oc, 0, 0);
	    fprintf(stderr,
		    "WARNING: An Ogg stream (serial %08x) with no header was\n"
		    "  encountered in %s.\n"
		    "  All data from this stream and any further such streams\n"
		    "  will be written to %s.\n\n%s",
		    ogg_page_serialno(&og), filename,
		    junk->filename, broken_ogg);
	  }
	  output_page_write(junk, &og);
	  continue;
	}

	/* this wasn't junk */
	chain_state=2;
      }

      /* we have a stream, write it to output file */
      output_page_write(stream->op, &og);

      if(ogg_page_eos(&og)){
	/* last page of stream */
	output_ctrl_output_free(&oc, stream->op->id);
	stream_ctrl_stream_free(&sc, stream->serial);

	if(sc.streams_used==0){
	  if(!unchain)
	    /* don't go on to get another link and beware of the evil goto */
	    goto doublebreak;

	  chain_state=0;
	  group_c=0;
	}
      }

    }
  }

 doublebreak:

  ogg_sync_clear(&oy);

  if(sc.streams_used){
    fprintf(stderr,
	    "WARNING: %d logical stream(s) remain unclosed at the end of %s.\n"
	    "  The stream may have been trunctated in an incomplete transfer\n"
	    "  or an aborted encoding process.\n\n%s",
	    sc.streams_used, filename, broken_ogg);
  }

  free(dirname);
  free(filename);

  output_ctrl_free(&oc);
  stream_ctrl_free(&sc);

  fclose(infile);

  return 1;
}

int main(int argc, char **argv)
{
  int c;
  opterr=0; /* Disable getopt error messages.  */
  while((c=getopt_long(argc,argv,optstring,options,NULL))!=EOF){
    switch(c){
    case 'c':
      ungroup=0;
      break;
    case 'g':
      unchain=0;
      break;
    case 'o':
      if(outdir!=NULL)free(outdir);
      outdir=xmalloc(strlen(optarg)+1);
      strcpy(outdir, optarg);
      break;
    case 'v':
      printf("oggsplit %d.%d.%d\n", VERSION_MAJOR, VERSION_MINOR, VERSION_MICRO);
      exit(0);
      break;
    case 'h':
    default:
      usage();
      break;
    }
  }

  /* if no files given, bye bye bye */
  if(optind==argc)
    usage();

  while(optind<argc){
    process_file(argv[optind++]);
    printf("\n");
  }

  printf("Done\n");

  if(outdir!=NULL)free(outdir);

  return 0;
}
